/* $Id: hdreg.h,v 1.4 1998/05/08 21:05:26 davem Exp $
 *
 *  linux/include/asm-mips/hdreg.h
 *
 *  Copyright (C) 1994-1996  Linus Torvalds & authors
 */

/*
 *  This file contains the MIPS architecture specific IDE code.
 */

#ifndef __ASM_MIPS_HDREG_H
#define __ASM_MIPS_HDREG_H

typedef unsigned short ide_ioreg_t;

#endif /* __ASM_MIPS_HDREG_H */

