#ifndef DECTYPES
#define	DECTYPES

#define DS2100_3100	1	/* DS2100/3100	Pmax		*/
#define DS5000_200	2	/* DS5000/200	3max		*/
#define DS5000_1XX	3	/* DS5000/1xx	kmin		*/
#define DS5000_2X0	4	/* DS5000/2x0	3max+		*/
#define DS5800		5	/* DS5800	Isis		*/
#define DS5400		6	/* DS5400	MIPSfair	*/
#define DS5000_XX	7	/* DS5000/xx	maxine		*/
#define DS5500		11	/* DS5500	MIPSfair-2	*/
#define DS5100		12	/* DS5100	MIPSmate	*/

#endif
