#ifndef __LINUX_X25_NETFILTER_H
#define __LINUX_X25_NETFILTER_H

/* X25-specific defines for netfilter.  Complete me sometime.
 * (C)1998 Rusty Russell -- This code is GPL.
 */

#include <linux/netfilter.h>

/* Hooks */
#define NF_X25_INPUT	0
#define NF_X25_FORWARD	1
#define NF_X25_OUTPUT	2

#endif /*__LINUX_X25_NETFILTER_H*/
