
/* this file generated on Thu Oct 24 11:42:37 GMT 1996  */

static char *action_names[] = { 
    "NOP"           ,"ADM1"          ,"ADM2"          ,"ADM3"          ,
    "ADM4"          ,"ADM5"          ,"CONN2"         ,"CONN3"         ,
    "CONN4"         ,"CONN5"         ,"RESWAIT1"      ,"RESWAIT2"      ,
    "RESWAIT3"      ,"RESWAIT4"      ,"RESWAIT5"      ,"RESWAIT6"      ,
    "RESWAIT7"      ,"RESWAIT8"      ,"RESCHK1"       ,"RESCHK2"       ,
    "RESCHK3"       ,"RESCHK4"       ,"RESCHK5"       ,"RESCHK6"       ,
    "SETUP1"        ,"SETUP2"        ,"SETUP3"        ,"SETUP4"        ,
    "SETUP5"        ,"SETUP6"        ,"SETUP7"        ,"SETUP8"        ,
    "RESET1"        ,"RESET2"        ,"RESET3"        ,"RESET4"        ,
    "RESET5"        ,"RESET6"        ,"RESET7"        ,"RESET8"        ,
    "D_CONN1"       ,"D_CONN2"       ,"D_CONN3"       ,"D_CONN4"       ,
    "D_CONN5"       ,"D_CONN6"       ,"D_CONN7"       ,"ERR1"          ,
    "ERR2"          ,"ERR3"          ,"ERR4"          ,"ERR5"          ,
    "ERR6"          ,"ERR7"          ,"ERR8"          ,"SH1"           ,
    "SH2"           ,"SH3"           ,"SH4"           ,"SH5"           ,
    "SH6"           ,"SH7"           ,"SH8"           ,"SH9"           ,
    "SH10"          ,"SH11"          ,"NORMAL1"       ,"NORMAL2"       ,
    "NORMAL3"       ,"NORMAL4"       ,"NORMAL5"       ,"NORMAL6"       ,
    "NORMAL7"       ,"NORMAL8A"      ,"NORMAL8B"      ,"NORMAL9"       ,
    "NORMAL10"      ,"NORMAL11"      ,"NORMAL12"      ,"NORMAL13"      ,
    "NORMAL14"      ,"NORMAL15"      ,"NORMAL16"      ,"NORMAL17"      ,
    "NORMAL18"      ,"NORMAL19"      ,"NORMAL20"      ,"BUSY1"         ,
    "BUSY2"         ,"BUSY3"         ,"BUSY4"         ,"BUSY5"         ,
    "BUSY6"         ,"BUSY7"         ,"BUSY8"         ,"BUSY9"         ,
    "BUSY10"        ,"BUSY11"        ,"BUSY12"        ,"BUSY13"        ,
    "BUSY14"        ,"BUSY15"        ,"BUSY16"        ,"BUSY17"        ,
    "BUSY18"        ,"BUSY19"        ,"BUSY20"        ,"BUSY21"        ,
    "BUSY22"        ,"BUSY23"        ,"BUSY24"        ,"BUSY25"        ,
    "BUSY26"        ,"REJECT1"       ,"REJECT2"       ,"REJECT3"       ,
    "REJECT4"       ,"REJECT5"       ,"REJECT6"       ,"REJECT7"       ,
    "REJECT8"       ,"REJECT9"       ,"REJECT10"      ,"REJECT11"      ,
    "REJECT12"      ,"REJECT13"      ,"REJECT14"      ,"REJECT15"      ,
    "REJECT16"      ,"REJECT17"      ,"REJECT18"      ,"REJECT19"      ,
    "REJECT20"      ,"AWAIT1"        ,"AWAIT2"        ,"AWAIT3"        ,
    "AWAIT4"        ,"AWAIT5"        ,"AWAIT6"        ,"AWAIT7"        ,
    "AWAIT8"        ,"AWAIT9"        ,"AWAIT10"       ,"AWAIT11"       ,
    "AWAIT12"       ,"AWAIT13"       ,"AWAIT14"       ,"AWAIT_BUSY1"   ,
    "AWAIT_BUSY2"   ,"AWAIT_BUSY3"   ,"AWAIT_BUSY4"   ,"AWAIT_BUSY5"   ,
    "AWAIT_BUSY6"   ,"AWAIT_BUSY7"   ,"AWAIT_BUSY8"   ,"AWAIT_BUSY9"   ,
    "AWAIT_BUSY10"  ,"AWAIT_BUSY11"  ,"AWAIT_BUSY12"  ,"AWAIT_BUSY13"  ,
    "AWAIT_BUSY14"  ,"AWAIT_BUSY15"  ,"AWAIT_BUSY16"  ,"AWAIT_REJECT1" ,
    "AWAIT_REJECT2" ,"AWAIT_REJECT3" ,"AWAIT_REJECT4" ,"AWAIT_REJECT5" ,
    "AWAIT_REJECT6" ,"AWAIT_REJECT7" ,"AWAIT_REJECT8" ,"AWAIT_REJECT9" ,
    "AWAIT_REJECT10" ,"AWAIT_REJECT11" ,"AWAIT_REJECT12" ,"AWAIT_REJECT13" ,
    	  0
};

