#ifndef _ASMARM_VT_H
#define _ASMARM_VT_H

#define VT_GETSCRINFO	0x56FD	/* get screen info */
#define VT_GETPALETTE	0x56FE  /* get palette */
#define VT_SETPALETTE	0x56FF  /* set palette */

#endif /* _ASMARM_VT_H */
